import sys

import menagerie
from . import new_project, generate

if __name__ == '__main__':
    if sys.argv[1] == 'new-project':
        try:
            new_project(sys.argv[2])
        except IndexError:
            new_project("NewProject")
    elif sys.argv[1] == 'generate':
        generate(*sys.argv)
    elif sys.argv[1] == 'version':
        print(f"Menagerie Version: {menagerie.__version__}")
    else:
        print(menagerie.__help__)



from pathlib import Path

from menagerie.Items.Managers.AbstractManager import AbstractManager
from menagerie.Items.Managers.StaticManager import StaticManager
from menagerie.Items.Meta.MetaItem import MetaItem
from menagerie.Items.Meta.XMLMetaItem import XMLMetaItem
from menagerie.Items.Meta.JSONMetaItem import JSONMetaItem


class MetaManager(AbstractManager):

    item_types = (MetaItem, XMLMetaItem)
    root_dir = ''

    def find(self):
        pass

    def initialize(self):
        self.items = (
            XMLMetaItem(self, Path('sitemap.jinja2'), 'xml'),
            XMLMetaItem(self, Path('browserconfig.jinja2'), 'xml'),
            JSONMetaItem(self, Path('manifest.jinja2'), 'webmanifest'),
            MetaItem(self, Path('robots.jinja2'), 'txt'),
        )

    def generate(self):
        self.base_env.globals['pages'] = self.gen.shared_info['pages']
        self.base_env.filters['static'] = self.gen.shared_info['static_filter']
        for item in self.items:
            item.generate()



from json import load, loads
from json.decoder import JSONDecodeError
from os import getenv
from pathlib import Path

import importlib.resources as pkg_resources
from jsonschema import validate
from jsonschema.exceptions import ValidationError
from ndicts.ndicts import NestedDict

import menagerie
from menagerie.utils.logger import Logger

__all__ = (
    'Settings',
    'setup_settings'
)

Settings = NestedDict({
    'config_path': None,
    'out_dir': None,
    'content_dir': None,
    'url_prefix': None,
    'default-toc': True,
    'base_url': '',
    'cache_enabled': True,
    'log_level': "Info",
    'themes': {
        'bootstrap': "https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/css/bootstrap.min.css",
        'highlight_js': "https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.4.0/styles/default.min.css",
        'navbar_color': "light"
    },
    'brand': {
        'app_name': "My App",
        'short_name': "My App",
        'favicon_folder': 'images/fav/',
        'navbar_icon': None,
        'navbar_icon_size': (100, 100),
        'socials': [],
        'show_gen_date': True,
        'footer': {
            'show_gen_time': True,
            'show_back_to_top': True,
            'show_made_with': True,
            'links': []
        },
        'meta': {
            'description': "",
            'keywords': [],
            'categories': [],
            'image': "",
            'image_alt': "Logo",
            'theme_color': "#333333",
            'bg_color': "#ffffff"
        }
    },
    'styles': {
        'base': None,
        'schema': None
    },
    'paths': {
        'out': "out/",
        'content': 'content/',
        'pages': 'pages/',
        'static': 'static/',
        'cache': '.m_cache/'
    },
    'minify':
        {
            'html': True,
            'css': True,
            'js': True,
            'xml': True,
            'json': True
        }
})


def setup_settings(config_path: Path):
    """
        Loads settings from a given config file and validates it against the config schema

        :param config_path: The path to the config file to load
        :type config_path: Path 
    """

    schema = loads(pkg_resources.read_text(menagerie, 'config_schema.json'))

    try:
        config = load(config_path.open(mode='r', encoding='utf-8'))
        validate(config, schema)
        Settings.update(NestedDict(config))
        Settings['out_dir'] = Path(Settings['paths', 'out'])
        Settings['content_dir'] = Path(Settings['paths', 'content'])
        Settings['url_prefix'] = getenv("URL_PREFIX", "")
        Settings['config_path'] = config_path
        Logger.update_level_from_string(Settings['log_level'])
    except FileNotFoundError:
        Logger.log_error(f"Can't find config file: `{config_path.as_posix()}`")
    except JSONDecodeError as e:
        Logger.log_error(f"Error parsing config: {e}")
    except ValidationError as e:
        Logger.log_error(f"Invalid config: {e.message}")



import os
import shutil
import importlib.resources as pkg_resources
import sys
from pathlib import Path

import menagerie


def new_project(name: str):
    with pkg_resources.path(menagerie, 'project_template') as template_path:
        new_path = Path(os.getcwd(), name)
        shutil.copytree(template_path, new_path)



from pathlib import Path

import importlib.resources as pkg_resources

from menagerie.Items.AbstractItem import AbstractItem
from menagerie.base_templates import meta_files


class MetaItem(AbstractItem):

    should_cache = False
    byte_mode = False
    extensions = ('*',)

    def __init__(self, manager, path: Path, out_ext):
        self.out_extension = out_ext
        super(MetaItem, self).__init__(manager, path)

    def get_content(self) -> str | bytes:
        return pkg_resources.read_text(meta_files, self.in_path)

    def initialize(self) -> None:
        pass

    def transform(self, content: str) -> str:
        return self.manager.base_env.from_string(content).render()



from htmlmin import minify

from menagerie.Items.MinifiedItemMixin import MinifiedItemMixin
from menagerie.Items.Meta.MetaItem import MetaItem
from menagerie.Items.Pages.AbstractPage import MINIFY_SETTINGS


class XMLMetaItem(MinifiedItemMixin, MetaItem):

    out_extension = 'xml'
    minify_key = 'xml'

    def minify(self, content: str) -> str:
        return minify(content, **MINIFY_SETTINGS)




from json_minify import json_minify

from menagerie.Items.Meta.MetaItem import MetaItem
from menagerie.Items.MinifiedItemMixin import MinifiedItemMixin


class JSONMetaItem(MinifiedItemMixin, MetaItem):

    minify_key = 'json'

    def minify(self, content: str) -> str:
        return json_minify(content)



import re
from typing import AnyStr

from jinja2 import UndefinedError
from xmlschema import XsdElement, XMLSchema as XMLDataSchema

from menagerie.Items.Pages.AbstractPage import AbstractPage


def children(group):
    child = [child for child in group if child.__class__.__name__ == "XsdElement"]
    for child_list in [children(inner_group) for inner_group in group if inner_group.__class__.__name__ == "XsdGroup"]:
        child += child_list
    return child


def ancestry(element):
    if element.parent is None:
        return [element.name]
    else:
        if element.name is None:
            return ancestry(element.parent)
        else:
            return [element.name] + ancestry(element.parent)


NCNAME_PATTERN = re.compile(r'^[^\d\W][\w.\-]*$')


class XMLSchema(AbstractPage):

    def load_metadata(self) -> None:
        for match in re.findall(r"<!--~(.*?)~-->", self.get_content(), re.MULTILINE):
            seperated = match.strip().split(':')
            self.meta[seperated[0].lower().replace('-', '_')] = seperated[1]
        self.meta['render_toc'] = False

    base_template = 'schema_templates/schema_template.jinja2'
    extensions = ('xsd', 'xml')

    def children(self, group):
        return children(group)

    def id_path(self, element):
        return '-'.join(reversed(ancestry(element)))

    def split(self, string, delim):
        return string.split(delim)

    def get_desc(self, element: XsdElement):
        try:
            return str(element.annotation.documentation[0].text.strip())
        except (UndefinedError, AttributeError):
            return ""

    def get_name(self, obj, unnamed='none'):
        try:
            name = obj.local_name
        except AttributeError:
            try:
                obj = obj.name
            except AttributeError:
                pass
            if not isinstance(obj, str):
                return unnamed
            try:
                if obj[0] == '{':
                    _, name = obj.split('}')
                elif ':' in obj:
                    prefix, name = obj.split(':')
                    if NCNAME_PATTERN.match(prefix) is None:
                        return ''
                else:
                    name = obj
            except (IndexError, ValueError):
                return ''
        else:
            if not isinstance(name, str):
                return ''
        return name.replace('.', '_').replace('-', '_')

    def occurs_text(self, occurs):
        words = {
            0: "Zero",
            1: "One",
            None: "Many"
        }
        return "Appears " + words[occurs[0]] + " To " + words[occurs[1]] + " " + ("Time" if occurs[1] == 1 else "Times")

    def inner_render(self, content: AnyStr) -> None:
        self.manager.base_env.filters.update({
            'children': self.children,
            'id_path': self.id_path,
            'split': self.split,
            'get_desc': self.get_desc,
            'get_name': self.get_name,
            'occurs_text': self.occurs_text
        })
        schema = XMLDataSchema(content)
        return self.manager.base_env.get_template('schema_templates/xml_schema.jinja2').render(page=self, schema=schema)




from abc import ABC

from menagerie.Items.AbstractItem import AbstractItem

__all__ = ('MinifiedItemMixin',)


class MinifiedItemMixin(AbstractItem, ABC):
    byte_mode = False
    minify_key: str = None

    def minify(self, content: str) -> str:
        raise NotImplementedError()

    def save(self, new_content: str) -> None:
        if self.manager.gen.settings['minify'][self.minify_key] is True:
            super(MinifiedItemMixin, self).save(self.minify(new_content))
        else:
            super(MinifiedItemMixin, self).save(new_content)



from json import loads, load, JSONDecodeError
from pathlib import Path
import importlib.resources as pkg_resources

from jsonschema.exceptions import ValidationError
from jsonschema.validators import validate
from markdown import Markdown

import menagerie
from menagerie.Items.Managers.AbstractManager import AbstractManager
from menagerie.Items.Pages.AbstractPage import AbstractPage, pretty_title
from menagerie.Items.Pages.HTMLPage import HTMLPage
from menagerie.Items.Pages.MDPage import MDPage
from menagerie.Items.Pages.Schema.JSONSchema import JSONSchema
from menagerie.Items.Pages.Schema.XMLSchema import XMLSchema
from jinja2 import Environment, FileSystemLoader

from menagerie.utils.logger import Logger


class NavItem:

    def __init__(self, content: str | list['NavItem'], is_folder: bool, meta: dict[str, object]):
        self.is_folder = is_folder
        self.titles = []
        self.meta = meta
        if is_folder:
            self.content = []
        else:
            self.content = content

    def add_item(self, nav_item: 'NavItem') -> None:
        if self.is_folder:
            self.content.append(nav_item)

    def iter_items(self) -> None:
        for page in sorted(self.content, key=lambda p: p.meta['sort_priority'], reverse=True):
            yield page

    def setup_titles(self) -> None:
        if self.is_folder:
            self.titles = [str(p.meta['title']).lower() for p in self.content]


class PageManager(AbstractManager):
    root_dir = 'pages'
    item_types: AbstractPage = (MDPage, HTMLPage, XMLSchema, JSONSchema)
    context: dict[str, object] = {

    }

    env: Environment = None
    router: dict[str, str] = {}

    def __init__(self, site_info) -> None:
        super(PageManager, self).__init__(site_info)
        self.env = Environment(loader=FileSystemLoader(self.gen.settings['content_dir']))
        self.root_dir = Path(self.gen.settings['paths', 'pages'])
        self.group_schema = loads(pkg_resources.read_text(menagerie, 'folder_schema.json', encoding='utf-8'))

    def find(self):
        super(PageManager, self).find()

    def initialize(self):
        for item in self.items:
            item.initialize()
        self.router = {p.meta['title'].lower(): str(p.out_path.relative_to(self.gen.settings['out_dir']).as_posix()) for p in self.items}
        self.gen.shared_info['pages'] = self.items
        self.filter_md = Markdown(extensions=['extra'], output_format='html5')

    def route(self, title: str) -> str:
        return self.gen.settings['url_prefix'] + self.router.get(title.lower(), "#")

    def make_nav_item_from_page(self, page: AbstractPage) -> NavItem:
        return NavItem(self.route(page.meta['title']), False, page.meta)

    def generate_nav_items(self) -> list[NavItem]:
        root_pages = []
        groups: dict[str, NavItem] = {}
        for page in self.items:
            if page.in_path.parent.name != '':
                if page.in_path.parent.name in groups.keys():
                    groups[page.in_path.parent.name].add_item(self.make_nav_item_from_page(page))
                else:
                    group_config_path = Path(page.get_path_to_open().parent, '_folder.json')
                    group_meta = {
                        'title': pretty_title(group_config_path.parent.name),
                        'sort_priority': 30
                    }
                    if group_config_path.exists():
                        try:
                            group_config = load(group_config_path)
                            validate(group_config, self.group_schema)
                            group_meta.update(group_config)
                        except JSONDecodeError as jde:
                            Logger.log_error(f"Couldn't parse config for {group_config_path.name} : {str(jde)}")
                        except ValidationError as ve:
                            Logger.log_error(f"Invalid folder config for {group_config_path.name} : {str(ve)}")
                    new_group = NavItem([], True, group_meta)
                    new_group.add_item(self.make_nav_item_from_page(page))
                    groups[group_config_path.parent.name] = new_group
            else:
                new_page = self.make_nav_item_from_page(page)
                root_pages.append(new_page)
        for group in groups.values():
            group.setup_titles()
        return sorted(list(groups.values()) + root_pages, key=lambda n: n.meta['sort_priority'], reverse=True)

    def generate(self):
        self.gen.settings['out_dir'].mkdir(parents=True, exist_ok=True)
        filters = {
            'route': self.route,
            'static': self.gen.shared_info['static_filter'],
            'upper_first': lambda x:   x[0].upper() + x[1:],
            'simple_md': lambda md: self.filter_md.convert(md)
        }
        da_globals = {
            'nav_items': self.generate_nav_items(),
        }
        self.base_env.filters.update(filters)
        self.base_env.globals.update(da_globals)
        self.env.filters.update(self.base_env.filters)
        self.env.globals.update(self.base_env.globals)
        for item in self.items:
            item.generate()
